#!/usr/bin/env python3
# encoding: UTF-8

# This file is part of turberfield.
#
# Turberfield is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Turberfield is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with turberfield.  If not, see <http://www.gnu.org/licenses/>.

import asyncio
from collections import deque
import unittest
import warnings

from turberfield.dynamics.simulation import JobQueue
from turberfield.dynamics.simulation import Simulation
from turberfield.dynamics.simulation import Stop
from turberfield.dynamics.simulation import Tick


class FreshLoop(unittest.TestCase):

    def setUp(self):
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)
        self.buf = deque()

    def tearDown(self):
        loop = asyncio.get_event_loop()
        loop.stop()
        loop.close()

    @asyncio.coroutine
    def loopback(self, q, jq, priority):
        tick = None
        self.nItems = 0
        while type(tick) is not Stop:
            tick = yield from q.get()
            yield from jq.put(tick.t, priority, tick)
            self.nItems += 1
        else:
            return tick


class TestTicks(unittest.TestCase):

    def test_ticks_fill_interval(self):
        seq = list(Simulation.ticks(0, 60, 1))
        self.assertEqual(59, max(i.t for i in seq if isinstance(i, Tick)))

    def test_ticks_non_repeating(self):
        seq = list(Simulation.ticks(0, 60, 1))
        self.assertEqual(len(set(seq)), len(seq))

    def test_ticks_end_with_stop(self):
        seq = list(Simulation.ticks(0, 60, 1))
        self.assertTrue(isinstance(seq[-1], Stop))
        self.assertEqual(60, seq[-1].t)


class TestTasks(FreshLoop):

    def test_task_has_queue(self):
        loop = asyncio.get_event_loop()
        jq = JobQueue(loop=loop)
        tsk = Simulation.task(loop, self.loopback, jq)
        self.assertTrue(hasattr(tsk, "q"))


class TestSimulation(FreshLoop):

    def test_ticks(self):
        loop = asyncio.get_event_loop()
        ticks = Simulation.ticks(0, 6, 1)
        jq = JobQueue(loop=loop)
        tsk = Simulation.task(loop, self.loopback, jq)
        sim = Simulation([tsk], jq, self.buf)
        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("ignore")
            try:
                loop.run_until_complete(asyncio.wait_for(
                    sim.dispatch(ticks), loop=loop, timeout=1))
                self.assertEqual(7, self.nItems)
                self.assertEqual(6, len(self.buf))
            except TimeoutError as e:
                self.fail(e)
            finally:
                sim._clear()
                self.assertFalse(hasattr(sim, "buf"))
                self.assertFalse(hasattr(sim, "jq"))
                self.assertFalse(hasattr(sim, "tasks"))



#!/usr/bin/env python3
# encoding: UTF-8

# This file is part of turberfield.
#
# Turberfield is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Turberfield is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with turberfield.  If not, see <http://www.gnu.org/licenses/>.

import asyncio
from asyncio.queues import QueueEmpty
from collections import deque
from collections.abc import MutableSequence
from functools import singledispatch
import itertools
import unittest
import warnings

from turberfield.dynamics.jobqueue import JobQueue
from turberfield.dynamics.types import Stop
from turberfield.dynamics.types import Tick


__doc__ = """
Here's how you would set up and run a simulation:

.. code-block:: python

    # The asyncio module is standard in Python 3.4 and above
    loop = asyncio.get_event_loop()

    # Create a generator for the simulation 'tick-stream'
    ticks = Simulation.ticks(0, 10, 1)

    # More about the JobQueue later...
    jq = JobQueue(loop=loop)

    # Simulation will run only a single task; 'agent', defined below.
    tasks = [Simulation.task(loop, agent, jq)]

    # Initialise the simulation object
    sim = Simulation(tasks, jq)

    # Run the simulation until it finishes or times out.
    loop.run_until_complete(asyncio.wait_for(
        sim.dispatch(ticks), loop=loop, timeout=1))


What is a task? It's a function or method which listens for the passage of
time and which can ask for things to happen in the future.
Here's what the task ``agent`` might look like:

.. code-block:: python

    @asyncio.coroutine
    def agent(q, jq, priority):
        tick = None
        while type(tick) is not Stop:
            tick = yield from q.get()
            # Do stuff...
            msg = "For the attention of..."
            yield from jq.put(tick.t + 1, priority, msg)
        else:
            return tick

``Agent`` is a coroutine. To work with the Turberfield simulator, it must
have these three positional parameters; `q`, `jq`, and `priority`:

    q
        Your task gets its :py:class:`ticks <turberfield.dynamics.types.Tick>`
        from this object.
    jq
        This is a :py:class:`~turberfield.dynamics.jobqueue.JobQueue` which
        your task uses to post messages.

    priority
        This is assigned to your task by the Turberfield simulator as
        a unique integer, thereby eliminating contention with other coroutines.

"""


@singledispatch
def message_handler(msg, *args, **kwargs):
    """
    The simulation calls this function, passing to it every message
    which pops up in the job queue.

    This function has been wrapped with the ``functools.singledispatch``
    decorator.

    So, to register your own supplied function as a handler for a particular
    class of message, make a call like the following::

        message_handler.register(type(msg), custom_handler)

    """
    warnings.warn("No handler found for {}".format(msg))
    rv = asyncio.Future()
    rv.set_result(msg)
    return rv


class Simulation:
    """
    The simulator in Turberfield is a shareable object. You need to
    define it once in your main module::

        sim = Simulation(tasks, jq, buf)

    Elsewhere, you can grab a readable copy like this::

        sim = Simulation(None)

    tasks
        A list of task objects. Each should be the result of a call
        to :py:func:`Simulation.task\
<turberfield.dynamics.simulation.Simulation.task>`.
    jq (optional)
        A :py:class:`~turberfield.dynamics.jobqueue.JobQueue` object.
        The simulation will get its messages here.
    buf (optional)
        A container which will record jobs as they are processed by the
        simulation. This can be useful for debugging purposes. If you
        omit this argument, the Simulation object will create its own
        (small) buffer.

    """

    _shared_state = {}

    @staticmethod
    def task(loop, fn, jq, priority=1, *args, **kwargs):
        """
        Create a `asyncio.Task` for `fn` (which must be a coroutine).
        The returned object is decorated with attributes necessary for
        use with a :py:class:`~turberfield.dynamics.simulation.Simulation`
        object. `jq` should be a
        :py:class:`~turberfield.dynamics.jobqueue.JobQueue` object common
        to all the tasks in your simulation.

        """
        q = asyncio.Queue(loop=loop)
        coro = fn(q, jq, priority, *args, **kwargs)
        rv = asyncio.Task(coro, loop=loop)
        rv.q = q
        rv.priority = priority
        return rv

    @staticmethod
    def ticks(start, end, interval):
        """
        Returns a generator of
        :py:class:`Ticks <turberfield.dynamics.types.Tick>` which will
        end with a
        :py:class:`Stop <turberfield.dynamics.types.Stop>`.

        """
        return itertools.chain(
            (Tick(i, 0) for i in range(start, end, interval)),
            [Stop(end, 0)])

    def __init__(self, tasks, jq=None, buf=None):
        self.__dict__ = self._shared_state
        if jq and not hasattr(self, "buf"):
            self.jq = jq
            self.buf = buf if isinstance(buf, deque) else deque(maxlen=32)
        if isinstance(tasks, list):
            if not hasattr(self, "jq"):
                warnings.warn("Simulation lacks a job queue")
            self.tasks = tasks

    def _clear(self):
        del self.jq
        del self.tasks
        del self.buf

    @asyncio.coroutine
    def dispatch(self, ticks):
        """
        The coroutine which drives the simulation loop.

        ticks
            An iterable of
            :py:class:`Ticks <turberfield.dynamics.types.Tick>` which
            ends with a
            :py:class:`Stop <turberfield.dynamics.types.Stop>`.

        """
        for tick in ticks:
            yield from asyncio.sleep(0)
            for tsk in (i for i in self.tasks if not i.done()):
                try:
                    yield from tsk.q.put(tick)
                except AttributeError as e:
                    warnings.warning("Task {} has no queue".format(tsk))
                    continue

            while True:
                try:
                    job, msg = self.jq.get_nowait()
                except QueueEmpty:
                    break
                else:
                    if tick.t < job.t:
                        yield from self.jq.put(job.t, job.priority, msg)
                        break
                    else:
                        self.buf.append((job, msg))
                        rv = yield from message_handler(
                            msg, Tick(tick.t, job.priority), self.jq)
        else:
            return tick



#!/usr/bin/env python
# encoding: UTF-8

import ast
import os.path

from setuptools import setup


try:
    # Includes bzr revsion number
    from turberfield.dynamics.about import version
except ImportError:
    try:
        # For setup.py install
        from turberfield.dynamics import __version__ as version
    except ImportError:
        # For pip installations
        version = str(ast.literal_eval(
            open(os.path.join(os.path.dirname(__file__),
                 "turberfield", "dynamics", "__init__.py"),
                 'r').read().split("=")[-1].strip()))

__doc__ = open(os.path.join(os.path.dirname(__file__), "README.txt"),
               'r').read()

setup(
    name="turberfield-dynamics",
    version=version,
    description="""
A package from the Turberfield project. Provides a simulation framework.
    """,
    author="D Haynes",
    author_email="tundish@thuswise.org",
    url="https://www.assembla.com/spaces/turberfield/messages",
    long_description=__doc__,
    classifiers=[
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3.4",
        "License :: OSI Approved :: GNU Affero General Public License v3"
        " or later (AGPLv3+)"
    ],
    namespace_packages=["turberfield"],
    packages=[
        "turberfield.dynamics",
        "turberfield.dynamics.test"
    ],
    package_data={
        "turberfield.dynamics": [],
        },
    install_requires=[
        "turberfield-common>=00.004",
        ],
    tests_require=[],
    entry_points={
        "console_scripts": [
        ],
    },
    zip_safe=False
)



#!/usr/bin/env python3
# encoding: UTF-8

# This file is part of turberfield.
#
# Turberfield is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Turberfield is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with turberfield.  If not, see <http://www.gnu.org/licenses/>.

import asyncio
from asyncio.queues import QueueEmpty
from collections import namedtuple


class JobQueue:
    """
    You create a JobQueue just as you would a standard Queue from the
    `asyncio` module.
    """

    Job = namedtuple("Job", ["t", "priority", "id"])

    def __init__(self, loop):
        self.loop = loop
        self.queue = asyncio.PriorityQueue(loop=loop)
        self.jobs = {}

    def get(self):
        """
        Returns a 2-tuple from the queue, consisting of:

        job
            A :py:class:`~turberfield.dynamics.JobQueue.Job` object,
            consisting of `(t, priority, id)`.
        msg
            The message delivered by the job.

        """
        job = yield from self.queue.get()
        return (job, self.jobs.pop(job.id))

    def get_nowait(self):
        job = self.queue.get_nowait()
        return (job, self.jobs.pop(job.id))

    def put(self, t, priority, msg):
        """
        Place a message on the queue, to be processed at time `t` with priority
        `priority`.

        """
        self.jobs[id(msg)] = msg
        yield from self.queue.put(JobQueue.Job(t, priority, id(msg)))

    def put_nowait(self, t, priority, msg):
        self.jobs[id(msg)] = msg
        self.queue.put_nowait(JobQueue.Job(t, priority, id(msg)))



#!/usr/bin/env python3.4
#   encoding: UTF-8

# This file is part of turberfield.
#
# Turberfield is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Turberfield is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with turberfield.  If not, see <http://www.gnu.org/licenses/>.

from collections import namedtuple

Tick = namedtuple("Tick", ["t", "priority"])
Tick.__doc__ = """A named tuple of {} to specify the time and priority order
of a message.""".format(Tick._fields)

Stop = namedtuple("Stop", ["t", "priority"])
Stop.__doc__ = """A special case of
:py:class:`~turberfield.dynamics.types.Tick`
which terminates a time sequence.""".format(Stop._fields)
